# -*- coding: utf-8 -*-
import os, warnings, copy, random, re
from laborejocore import cursor, lilypond, helper, pitch
from laborejocore.items import * #Internal, Playback, Lilypond, Start, End, PerformanceSignature, KeySignature, TimeSignature, Appending, LilypondDefinitions, Header, Chord, Rest, AlternateEnd, AlternateEndClose, RepeatClose, RepeatOpen, RepeatCloseOpen, Rest, Fine, Segno, GotoSegno, GotoCapo, GotoCoda, SpecialBarline, Clef, MultiMeasureRest
from string import ascii_letters

try:
    import smf
except:
    warnings.warn("libpysmf not found. Midi export and Playback broken. Not 'disabled', broken! You can use the rest of the program just fine.")

class Container(Internal, Playback, Lilypond):
    """Basic and smallest class to hold several items.
    Can be nested. Emulates list behaviour."""
    #TODO: Do not export containers as plain unfolded lilypond but put them in vars above the movement. Create %$$MUSICDEFINITIONS$$
    #This is the data structure that needs the performance features.
    def __init__(self):
        super(Container, self).__init__()
        self.container = [Start(self), End(self)]
        self.instanceCount = 0
        self._uniqueContainerName = self.generateNextUniqueContainerName()
        self.cursorWalk = True #The cursor walks through this as if the container content were single items. This is meant for direct access by the user. If he does not want to edit the container with the general cursor he can deactivate this. Meant for external edit-the-container modes or tools.
        self.unfold = True #For lilypond (and possible GUIs) unfold this on export. Midi will always be unfolded. This is False for effects and some substitutions. The trill will only be rendered as midi, never as lilypond, gui or analysis.
        self.group = "" #An optional group for various purposes.
        self.repeatPercent = 1 #numbers bigger than 1 will be unfolded as % repeat with lilyponds \repeat percent n {} syntax
        self.figures = self.chordsymbols = None #prevents typchecking


    def append(self, value):
        self.container.append(value)
    def __len__(self):
        return len(self.container)
    def __getitem__(self, i):
        return self.container[i]
    def __iter__(self):
        return iter(self.container)
    def __setitem__(self, item, value):
        self.container[item] = value
    def __delitem__(self, item):
        del self.container[item]

    def insert(self, i, item):
        self.container.insert(i, item)

    def flatten(self):
        """return a flattened version of self"""
        for elem in self.container:
            if type(elem) in (tuple, list):
                for i in flatten(elem):
                    yield i
            else:
                yield elem

    def generateNextUniqueContainerName(self):
        """Return an 8 letter string if someone forgot to choose a name for Track. Its guaranteed to be random @ creation time."""
        try: #is this the program start? The first track is generated before the workspace.
            session.workspace
            stringy = ''.join([random.choice('abcdefghijklmnoprstuvwyxzABCDEFGHIJKLMNOPRSTUVWXYZ') for i in range(8)])
            while stringy in session.workspace.uniqueContainerNames:
              stringy = ''.join([random.choice('abcdefghijklmnoprstuvwyxzABCDEFGHIJKLMNOPRSTUVWXYZ') for i in range(8)])
            session.workspace.uniqueContainerNames.append(stringy)
            return stringy
        except:
            return ''.join([random.choice('abcdefghijklmnoprstuvwyxzABCDEFGHIJKLMNOPRSTUVWXYZ') for i in range(8)])

    @property
    def uniqueContainerName(self):
        return self._uniqueContainerName

    @uniqueContainerName.setter
    def uniqueContainerName(self, value):
        value = "".join([ch for ch in value if ch in ascii_letters ]) #remove all non-chars to create valid lilypond syntax
        if value in session.workspace.uniqueContainerNames or value == "":
            return False
        else:
            self._uniqueContainerName = value
            return True

    @property
    def selected(self):
        return False

    @selected.setter
    def selected(self, value):
        return False

    @property
    def duration(self):
        """Complete Musical Duration of this container. Used to calculate sync and playback"""
        #[:-1] means without the End() item. Else we get an recursive loop since End().duration asks for this self.duration() function again.
        return sum([x.duration for x in self.container[:-1]])

    def isEmpty(self):
        """Has no duration"""
        try:
            next(x for x in self.container if x.duration)
            return False
        except StopIteration:
            return True

    def select(self, opt = False):
        pass
    def deselect(self, opt = False):
        pass
    #def select(self):
    #    """Select any item or container in a container. In the end it
    #    always results in selected items, not containers"""
    #    for x in self:
    #        x.select()

    #def deselect(self, deleteMe = False):
    #    for x in self:
    #        x.deselect(deleteMe)

    def recursiveParseWithTuplets(self, generateFunctionString):
        """Used for generate Lilypond data ("inner"), chordsymbols and
        figured bass. generateFunctionString is used to tell the
        item which function to use.
        It works with notes, figuredBass and
        chordsymbols for Chords/Items so recursiveParse is the same
        function for all these export steps"""
        returnList = []
        lastTuplet = None
        current = 0 #not needed logically, but technically the first test down here uses "current" before assignment.
        lilypondTupletLevel = 0 #For tuplet merging in Lilypond. Used to check if a new tuplet starts or if a tuplet is finished. Can be compared to chord.tuplets list length.
        for x in self:
            if x.tuplets and lilypondTupletLevel == len(x.tuplets) and (not x.tuplets == lastTuplet):
                #If an old tuplet ends and a new but differnent one start on the same level close and reopen the new one.
                returnList.append("}") #close old tuplet.
                returnList.append("\\times " + str(x.tuplets[current][0]) + "/" + str(x.tuplets[current][1]) +  "{") # Begin new tuplet
            if x.tuplets:
                lastTuplet = x.tuplets
            if x.tuplets and lilypondTupletLevel < len(x.tuplets):
                while not lilypondTupletLevel == len(x.tuplets):
                    lilypondTupletLevel += 1
                    current = lilypondTupletLevel -1
                    returnList.append("\\times " + str(x.tuplets[current][0]) + "/" + str(x.tuplets[current][1]) +  "{") # Begin a tuplet
            if lilypondTupletLevel > len(x.tuplets):
                while not lilypondTupletLevel == len(x.tuplets):
                    #returnList.append("}") #close tuplets.
                    session.workspace.waitForNextChordPre.append("}")

                    lilypondTupletLevel -= 1
            returnList.append(getattr(x, generateFunctionString)()) #the actual data gathering
        return " ".join(returnList)

    def generateInnerLilypond(self):
        """The string for recursive tuplets is used to choose a function
        which will be exported. It works with notes, figuredBass and
        chordsymbols for Chords/Items so recursiveParse is the same
        function for all these export steps"""
        if self.repeatPercent > 1:
            return "\\repeat percent " + str(self.repeatPercent) + " { " +  self.recursiveParseWithTuplets("exportLilypond") + " } "
        else:
            return self.recursiveParseWithTuplets("exportLilypond")

    def generateFiguredBass(self):
        """The string for recursive tuplets is used to choose a function
        which will be exported. It works with notes, figuredBass and
        chordsymbols for Chords/Items so recursiveParse is the same
        function for all these export steps"""
        if self.repeatPercent > 1:
            return "\\repeat percent " + str(self.repeatPercent) + " { " +  self.recursiveParseWithTuplets("generateFiguredBass") + " } "
        else:
            return self.recursiveParseWithTuplets("generateFiguredBass")

    def generateChordsymbols(self):
        """The string for recursive tuplets is used to choose a function
        which will be exported. It works with notes, figuredBass and
        chordsymbols for Chords/Items so recursiveParse is the same
        function for all these export steps"""
        if self.repeatPercent > 1:
            return "\\repeat percent " + str(self.repeatPercent) + " { " +  self.recursiveParseWithTuplets("generateChordsymbols") + " } "
        else:
            return self.recursiveParseWithTuplets("generateChordsymbols")

    def generateLilypond(self):
        """Export only the uniqueContainerName  which is the lilypond
        variable.
        But save the actual data in a temporary, non-saved,
        dictionary in session.workspace."""
        innerLy = self.generateInnerLilypond() #returns a string
        session.workspace.movement.nonTrackContainer[self.uniqueContainerName] = innerLy
        return " \\" + self.uniqueContainerName

    def generateSubstitutionLilypond(self):
        """Substitutions need the real string, not a containername"""
        return self.generateInnerLilypond()

    def transform(self, keysig, referenceBaseDuration=None, rootpitch =  None, scale = [None, None, None, None, None, None, None]):
        """Return a copy of itself, destructively transformed by
        the optional parameters.
        If none are given it returns just a deepcopy.
        Every object and the structure itself is new.

        rootpitch is a pitch which will be compared with 1420
        None means no change.
        1470 means all notes are shifted in the ratio
        as c' to d' (major 2nd up)

        referenceBaseDuration is used to calculate a ratio.
        A whole note duration and a whole note container duration sum
        will result in a factor 1 (no change).
        parameter duration 768/half note and container
        duration 1536/whole note will result in 0.5.
        Every note in the new.container will have its base duration
        multiplied with 0.5.

        scales positions list:
        The lowest pitch is considered the root note.
        Return a list of 8 positions. The first is the octave, a bool that
        shows if any higher octave variants of the root are in here.
        For all other positions, if present, return the difference to white.

        All other note scale positions/intervals force the position to
        a specific alteration/accidental.
        If an interval is given for a position in the scale then the
        original pitch is forgotten and a new one gets calculated with
        the intervalUp and the rootpitch.
        """
        new = copy.deepcopy(self)
        if not len(scale) == 7:
            raise ValueError("The given scale needs to have exactly 7 positions. From root to seventh. Use None to skip one")
            return new

        def filterCheck(item):
            typ = type(item)
            if typ in [KeySignature, Clef, TimeSignature, Container, Track, Movement, Appending, Start]: #End must stay in because .duration() excludes it with [:-1]
                return False
            else:
                return True
        filtered = list(filter(lambda chk: filterCheck(chk), new.container))
        new.container = filtered

        #Now all critical items are filtered out. Begin transformation process.
        if referenceBaseDuration:
            completeContainerDuration = new.duration #the complete duration of the container
            ratio = referenceBaseDuration / completeContainerDuration
        else:
            ratio = None

        for item in new.container[:-1]: #without End()
            if ratio and not type(item) is End:
                item.guessDuration(referenceBaseDuration, completeContainerDuration) #changes the items duration in place.

            if rootpitch: #this is not really optional. If somehow this is avoided (certainly not by chord.exportLilypond() or playback) we just get a plain substitution with the exact container content in C.
                for note in item.notelist:
                    scalePos = pitch.diatonicIndex(note.pitch)
                    #transpose the given note to the current keysig or do a real transposition, if there is an octave.
                    note.intervalAutomatic(1420, rootpitch)
                    if not scalePos == 0:
                        if scale[scalePos] or scale[scalePos] == 0:
                            note.toWhite()
                            note.pitch += scale[scalePos]
                        elif not scale[0]: #There was no octave in the reference scale. This means not real tranposition but shift.
                            note.toScale(keysig)
            else:
                warnings.warning("Container transform without root note. This is very uncommon.", self, self.uniqueContainerName)

        return new

    def saveState(self):
        pass

    def generateSave(self):
        """If in doubt return nothing, still compatible.
        Child classes should implement this function and return dict"""
        return {}

    def exportSave(self):
        """Generate the dict with whatever the child class gives us
        then add the Internal Values (durations) and return as tuplet
        (ClassName, childDict) which will be exported to Json in the end"""
        childDict =  self.generateSave()

        #Internal
        childDict["base"] = self.base
        childDict["tuplets"] = self.tuplets
        childDict["dots"] = self.dots
        childDict["scaleFactorNumerator"] = self.scaleFactorNumerator
        childDict["scaleFactorDenominator"] = self.scaleFactorDenominator
        #Playback
        childDict["split"] = self.split
        childDict["durationFactor"] = self.durationFactor
        childDict["instructionPre"] = self.instructionPre
        childDict["instrcutionPst"] = self.instructionPst
        #Lilypond
        childDict["lilypond"] = self.lilypond
        childDict["directivePre"] = self.directivePre
        childDict["directiveMid"] = self.directiveMid
        childDict["directivePst"] = self.directivePst

        exportList = []
        for item in self.container:
            exportList.append(item.exportSave())

        childDict["cursorWalk"] = self.cursorWalk
        childDict["group"] = self.group
        childDict["unfold"] = self.unfold
        childDict["_uniqueContainerName"] = self._uniqueContainerName
        childDict["container"] = exportList
        childDict["repeatPercent"] = self.repeatPercent

        return (type(self).__name__, childDict)

    #No export Playback here. Export playback is done by the cursor which goes transparently through all containers.

class Track(Container):
    """Medium sized container. Can hold containers or items. Only one keysig and timesig is valid at each time-position here.
    Attention: Testing for equality or not equal with tracks is done through the unique track name. tracks which have the same
    name are considered equal. This was done by overriding __eq__ and __ne__, see below."""
    #StaffGroups are normally Spanners. But they can be nested at will. So we have a situation here.
    ##1)Risk users forgetting to close or open a group
    ##2.Overly complex container situation with nested containers where multiple starts and endings are in the same container.
    ##1) was chosen.
    def __init__(self):
        super(Track, self).__init__()
        self.container = [Appending()] #no Start() in tracks
        self.staffType = "Staff" #exports to "\new Staff {c d e}"  #TODO: Why is a layout option in here? If staffype becomes a RhythmStaff etc. we need to elimate pitches.
        self.nrOfLines = 5
        self.mergeWithUpper = False # Is this Track a voice or a standalone staff? Default False = Standalone.
        self.longInstrumentName = ""  #long instrument name, not unique.
        self.shortInstrumentName = ""  #short instrument name, not unique.
        self.voicePreset = "" #These are the lilypond voice presets. "" means surpressed output. Can be voiceOne, voiceTwo, voiceThree, voiceFour or oneVoice. OneVoice is the explicit version of "None"
        self.trackDirectiveWith = {} # at least one entry here creates a \with {} block. Only one block is allowed per staff so we can't use the other directives.
        self.selectedItems = [] #a shortcut list to all selected Items.
        self.hideLilypond = False
        self.transposition = (1420, 1420)  #[0]=what is meant, [1]=what will be written in the pdf . Usually [0] stays 1420 and [1] changes. Exports as lilypond transpose for pdf generation. Sounds the same, probably looks different. initially no transposition. c' is c'.  In human words: "I notated in C and it sounds correct but for the pdf I want it to be notated one whole tone lower for a clarinet"
        self.lilypondStaffGroups = []
        self.lyrics = "" #empty lines get converted as new verses. this means a double \n[white]\n  with any whitespace in between removed.
        self.forceLyricsAbove = False
        self.verseCountCache = 0
        self.endingBar = '\\bar "|."'
        self.lilypondSize = "normal"  #normal, big, small, tiny.
        self.ambitus = False # Ambitus engraver

        #Simple SMF Data
        #self.smfPort does not exist. Threre is only one port in standard smf.
        self.smfChannel = [0,0]
        self.smfMute = False
        self.smfSolo = False
        self.smfPatch = 0
        self.smfBank = 0
        self.smfVolume = 99 #max is 127. You can get up a little now.
        self.smfPan = 63 #0-63, 64-127
        self.smfTransposition = 0 #Half Tones. negative or positive. This is just for midi export. If you export your Contrabass with -12 here it will sound correct but it will be still considered as written in Laborejo in the analysis. Simple rule: Type in what it is.

        #Jack Midi Data
        self.jackPort = ""  # syntax: "client:port"  for example "Linuxsampler:Violins"
        self.jackChannel = [0,0]
        #self.jackMute = False  #not mirrored. Mute is mute.
        #self.jackSolo = False  #not mirrored. Solo is solo.
        self.jackPatch = 0
        self.jackBank = 0
        self.jackVolume = 127 #usually the volume is done in an audio mixer or the sampler/synth.
        self.jackPan = 63
        self.jackTransposition = 0
        self.jackAudioReturnPortLeft = "" #for exporting functions if they want to connect the resulting audio from self.jackPort to a recording strip.
        self.jackAudioReturnPortRight = ""

    def __eq__(self, other):
        """Tracks are equal if the unique(!) name is the same."""
        return self._uniqueContainerName == other
    def __ne__(self, other):
        """Tracks are not equal if the unique(!) name is not the same."""
        return self._uniqueContainerName != other

    def select(self, registerInTrackSelectedItemsList = True):
        """Select any item or container in a container. In the end it
        always results in selected items, not containers"""
        cur = copy.copy(session.workspace.movement.cursors[session.workspace.movement.container.index(self)]) #only copy cursor, not the items.
        cur.go_to_head()
        cur.resetToDefault()
        for x in cur.iterate():
            x.select(registerInTrackSelectedItemsList)

    def deselect(self, deleteMe = False):
        """This list does not define what is selected and what not.
        Its just a shortcut for deselecting and counting"""
        #for x in self.selectedItems:
        #    x.deselect(False) # Do not delete by active track...
        #    session.workspace.movement.selection -= 1 #decrease the counter
        #self.selectedItems = [] #...delete all at once instead

        cur = copy.copy(session.workspace.movement.cursors[session.workspace.movement.container.index(self)]) #only copy cursor, not the items.
        cur.go_to_head()
        cur.resetToDefault()
        for x in cur.iterate():
            x.deselect(False) # Do not delete by active track...
            session.workspace.movement.selection -= 1 #decrease the counter
        self.selectedItems = [] #...delete all at once instead

    def generateLyrics(self):
        """Returns lists, each one is a verse.
        Also transforms \n\n to a new verse and
        numbers to \"123\" """

        escaped = re.sub(r'(\d+)', r'"\1"', self.lyrics)
        escaped = escaped.replace("__,", " \\extendComma") #a special command, you need to include inc_lyricsextender.ly
        escaped = escaped.replace("__!", " \\extendExclaim") #a special command, you need to include inc_lyricsextender.ly
        escaped = escaped.replace("<i>", " \\override LyricText #'font-shape = #'italic ")
        escaped = escaped.replace("<n>", " \\override LyricText #'font-shape = #'normal-text ")
        escaped = escaped.split("\n\n")
        self.verseCountCache = len(escaped)
        return escaped

    def getLyricsAsList(self):
        """Return each syllabe as List member"""
        return [verse.replace(" -- ", "-- ").split() for verse in self.lyrics.split("\n\n")]

    def generateWithBlock(self):
        size = {
                "":"", #in case of error or manual deletion
                "normal": "",
                "small": """fontSize = #-3 \override StaffSymbol #'staff-space = #(magstep -3)""",
                "big": """fontSize = #3 \override StaffSymbol #'staff-space = #(magstep 3)""",
                "tiny": """fontSize = #-7 \override VerticalAxisGroup #'minimum-Y-extent = #'(0 . 0) \override StaffSymbol #'staff-space = #(magstep -7)""",
        }[self.lilypondSize]

        ambitus = '\\consists "Ambitus_engraver"' if self.ambitus else ""
        instNames = self.generateLilypondStaffNames() if self.longInstrumentName or self.shortInstrumentName else ""

        if self.trackDirectiveWith or size or ambitus or instNames:  #excludes normal. good.
            return " \\with { " + size  + " ".join(self.trackDirectiveWith.values()  ) + ambitus + instNames + " } "
        else:
            return ""

    def generateLilypondStaffNames(self):
        """Return the name of the staff, short and long format"""
        #TODO: If merged voice the instrument names should be combined. Must work for several merged voices.
        """
        longName = ""
        shortName = ""
        if self.longInstrumentName: #Long name
            longName = "\\set " + self.staffType + ".instrumentName = #\"" + self.longInstrumentName + "\""  #TODO: maybe self.staffType is wrong here. Does it work with rhytmic staff?
        if self.shortInstrumentName: #Short name
            shortName = "\\set " + self.staffType + ".shortInstrumentName = #\"" + self.shortInstrumentName + "\""
        return longName + " " + shortName
        """

        #Long and short instrument names
        longName = " instrumentName = \"" + self.longInstrumentName + "\"  " if self.longInstrumentName else ""
        shortName = " shortInstrumentName = \"" + self.shortInstrumentName + "\" " if self.shortInstrumentName else ""
        return longName + shortName

    def generateLilypondVoice(self):
        """Generate the actual data, indenpendent of single voice or merged"""
        #session.workspace.lastDurationLyString = "" #reset last duration cache
        figuredBass = self._uniqueContainerName + "Continuo = \\figuremode { " + self.generateFiguredBass() + " }\n" if any([item.figures for item in self.flatten()]) else ""
        chordsymbols = self._uniqueContainerName + "Chordsymbols = \\new ChordNames \\chordmode { " + self.generateChordsymbols() + " }\n" if any([item.chordsymbols for item in self.flatten()]) else ""
        lyrics = "".join([self._uniqueContainerName + "Lyrics" + helper.num2word(verse[0]) + "= \\lyricmode { " + verse[1] + " }\n" for verse in enumerate(self.generateLyrics())]) if self.lyrics else "" #Generate one defition line for each verse
        innerLilypond = self.generateInnerLilypond()

        if self.uniqueContainerName.lower() == "master":
            override = "\\override Voice.Dots #'stencil = ##f \\override Voice.TimeSignature #'stencil = ##f \\override Voice.KeySignature#'stencil = ##f \\override Voice.Clef #'stencil = ##f \\override Voice.BarLine #'stencil = ##f  \\override Voice.NoteHead #'stencil = ##f  \\override Voice.Stem #'stencil = ##f \\override Voice.Rest  #'stencil = ##f \\override Voice.MultiMeasureRestNumber  #'stencil = ##f \\override Voice.MultiMeasureRestText  #'stencil = ##f \\override Voice.MultiMeasureRest  #'stencil = ##f  "
            return self._uniqueContainerName + "= { " + override  + innerLilypond +  "}\n" + figuredBass + chordsymbols + lyrics

        else: #normal track
            #Only Appending() in container or even broken (0 objects)?. Empty Track, add one skip to avoid broken lilypond!
            #This does not prevent empty pages if there are only clefs and such in the lilypond output. But at least that is not an error.
            fallback = " s4 " if len(self) <= 1 else ""

            #if there is no keysig at the track beginning force C Major so transpose will work correctly. C Major is the lilypond default.
            keyFallback =  "\\key c \\major " if not ( any(type(x) is KeySignature for x in self[:5])) else ""

            voicePreset = "\\" + self.voicePreset + " " if self.voicePreset else ""

            #If the music ends with a custom barline use this, if not add an ending bar.
            ending = "" if  re.search(r'(\\bar ".+"\s*)$', innerLilypond) else self.endingBar

            linesNr = "\\override Staff.StaffSymbol #'line-count = #" + str(self.nrOfLines) + " " if not self.nrOfLines == 5 else ""
            #return self._uniqueContainerName + "= { " + self.generateLilypondStaffNames() + linesNr + voicePreset + keyFallback + innerLilypond + fallback +  ending + "}\n" + figuredBass + chordsymbols + lyrics
            return self._uniqueContainerName + "= { " + linesNr + voicePreset + keyFallback + innerLilypond + fallback +  ending + "}\n" + figuredBass + chordsymbols + lyrics

    def generateLilypond(self):
        """Generate the line that it is used in \score << >> to arrange."""
        figuredBass = " \\new FiguredBass \\" + self._uniqueContainerName + "Continuo " if any([item.figures for item in self.flatten()]) else ""
        chordsymbols = "\\" + self._uniqueContainerName + "Chordsymbols " if any([item.chordsymbols for item in self.flatten()]) else ""
        #Lyrics
        lyricsAfter = lyricsBefore = ""
        if self.lyrics:
            lyrics = "".join(["\\new Lyrics = \"" + self._uniqueContainerName + "Lyrics" + helper.num2word(verseCounter) + "\" {s4} " for verseCounter in range(self.verseCountCache)])
            if self.forceLyricsAbove:
                lyricsBefore = lyrics
            else:
                lyricsAfter = lyrics

                figuredBass = " \\new FiguredBass \\" + self._uniqueContainerName + "Continuo " if any([item.figures for item in self.flatten()]) else ""
                chordsymbols = "\\" + self._uniqueContainerName + "Chordsymbols " if any([item.chordsymbols for item in self.flatten()]) else ""

        if self.uniqueContainerName.lower() == "master":
            return chordsymbols + lyricsBefore + " \\new Voice = \"" + self._uniqueContainerName + "\" " + "\\" + self._uniqueContainerName + " " + lyricsAfter +  figuredBass

        else: #normal track
            if self.hideLilypond:
                return ""
            else:
                if self.mergeWithUpper:
                    pre = "" # x is voice of another staff.
                    master = ""
                else:
                    master = session.workspace.movement.master.exportLilypond() if session.workspace.movement.master else ""
                    pre = " \\new " + self.staffType + ' = "' + self._uniqueContainerName + '_Staff" ' + self.generateWithBlock() + " << " #x is a stand alone staff beginning

                voiceMid = " ".join(self.directiveMid.values())

                if self.transposition == (1420, 1420):
                    transpose = ""
                else:
                    transpose = " \\transpose " + lilypond.pitch2ly[self.transposition[0]] + " " + lilypond.pitch2ly[self.transposition[1]] + " "
                return chordsymbols + lyricsBefore + pre  + voiceMid + master + voiceMid + " \\new Voice = \"" + self._uniqueContainerName + "\" " + transpose  + "\\" + self._uniqueContainerName + " " + lyricsAfter +  figuredBass


    def generateSave(self):
        childDict = {}
        childDict["staffType"] = self.staffType
        childDict["nrOfLines"] = self.nrOfLines
        childDict["mergeWithUpper"] = self.mergeWithUpper
        childDict["longInstrumentName"] = self.longInstrumentName
        childDict["shortInstrumentName"] = self.shortInstrumentName
        childDict["trackDirectiveWith"] = self.trackDirectiveWith
        childDict["smfChannel"] = self.smfChannel
        #childDict["selectedItems"] = self.selectedItems Selection is not saved
        childDict["smfMute"] = self.smfMute
        childDict["smfSolo"] = self.smfSolo
        childDict["smfPatch"] = self.smfPatch
        childDict["smfBank"] = self.smfBank
        childDict["smfVolume"] = self.smfVolume
        childDict["smfPan"] = self.smfPan
        childDict["transposition"] = self.transposition
        childDict["smfTransposition"] = self.smfTransposition
        childDict["hideLilypond"] = self.hideLilypond
        childDict["voicePreset"] = self.voicePreset
        childDict["lyrics"] = self.lyrics
        childDict["forceLyricsAbove"] = self.forceLyricsAbove
        childDict["lilypondSize"] = self.lilypondSize
        childDict["ambitus"] = self.ambitus
        childDict["jackPort"] = self.jackPort
        childDict["jackChannel"] = self.jackChannel
        childDict["jackPatch"] = self.jackPatch
        childDict["jackBank"] = self.jackBank
        childDict["jackVolume"] = self.jackVolume
        childDict["jackPan"] = self.jackPan
        childDict["jackTransposition"] = self.jackTransposition
        childDict["jackAudioReturnPortLeft"] = self.jackAudioReturnPortLeft
        childDict["jackAudioReturnPortRight"] = self.jackAudioReturnPortRight
        return childDict


class Movement(Container):
    """The biggest and highest level container. Only Tracks in here
     Movement also holds all higher level cursor functions because all cursors in a movement move are in sync.
     Tracknumbers and cursor numbers are in sync if only internal functions are used to create and delete Tracks.
     Tracks are exported and parsed in the order they are created or swapped. Container index is export position."""
    #TODO: Maybe its wrong to have the complex movement cursors in here and not in its own class. Only functions which involve self.container need to be in here.
    def __init__(self):
        super(Movement, self).__init__()
        self.activeCursor = None #A cursor instance: What is the currently active cursor (aka. which Track is active). New cursors get added through addTrack
        self.masterCursor = None #A cursor instance. An instance is only added during exporting (playback, for example). During work the mastertrack and its cursors are just normal tracks. Not saved.
        self.cursors =  [] # All cursors instances. One for each Track.
        self.container = [] #No appending here.
        self.nonTrackContainers = {} #All containers used in this movement. For a quick access. key is the container instance, value is the instance count.
        self.substitutions = {}  #Dynamic. Has Container() in it. This is saved and loaded with the movement file. A gui should offer add/remove/edit options for the content.
        self.selection = 0 #0 means no selection.
        self.subtext = "" #Subtext, word wrapped under the ly score, movement. Goes directly to Lilypond export.
        self.defaultPerformanceSignature = PerformanceSignature()
        self.nonTrackContainer = {} #a temporary storage for non-track containers during ly export. uniqueContainerName is the key, a lilypond-string is the value.
        self.master = None #A track. An instance is only added during exporting (playback, for example). During work the mastertrack and its cursors are just normal tracks. Not saved.
        self.stateRegistry = [] #undo stack.
        self.oldSnapshots = [] #for redo. Makes only sense if there is no data change between undo and redo. This gets cleaned by the api or other high level functions.
        self.lyFontSize = 20 #Can be used in a Lilypond export template.
        self.addTrack() #Finally add a track. Each movement needs at least one.

    """
    def snapshotCreate(self):
        #Deprecated
        #A snapshot is a tuplet (identifier, pickled State)
        snap = pickle.dumps((self.cursors, self.activeCursor, self.container))
        return (uuid.uuid1(), pickletools.optimize(snap)) #return the whole snapshot

    def snapshotSave(self):
        #Deprecated
        #The basic undo function. Save all cursors and the movement data on a stack.
        self.snapshots.append(self.snapshotCreate())
        return self.snapshots[-1][0] #return the uuid

    def snapshotLoad(self, snap):
        #Deprecated
        #Load a snapshot of all cursors and the movement data
        identifier = snap[0] #not used here.
        snap = pickle.loads(snap[1])
        self.cursors = snap[0]
        self.activeCursor= snap[1]
        self.container = snap[2]
        return identifier # return the loaded uuid
    """

    def select(self):
        """This is different from any other container because it does
        not rely on a cursor position anymore so item.select would not
        know where to place the selected item in self.selectedItems"""
        for x in self: #tracks!
            x.select(registerInTrackSelectedItemsList = False)
        #recreate the selectedItems per-track lists by copying the actual movement and flattening the tracks.
        for track in self.container:
            track.selectedItems = list(helper.flatten(track.container)) #flatten a copy(!) of the track container and make it the new selectedItems, overwriting all the old selected data.
            track.selectedItems.pop() #delete the last item in the list, which is the final appending position which got copied here because we did not used item.select().

    def deselect(self, opt= False):
        """Deselect any item."""
        for tr in self.container:
            tr.deselect(False)
        self.selection = 0

    #Synced Cursor Insert and Movement in all Tracks.
    def insertAtCursor(self, item):
        r = self.activeCursor.insert(item) #goes right afterwards so we need the other cursors to follow
        self.comeLittleCursors()
        return r

    def syncCursors(self):
        """After inserting or modifying items linked/container
        items in multiple tracks the tick positions are wrong.
        Update all tracks"""
        self.activeCursor.syncTickIndex()
        for x in [cur for cur in self.cursors if not cur is self.activeCursor]:
            x.go_to_head()
            x.resetToDefault() #Reset all counters and gatherers to 0.
        self.comeLittleCursors()

    def comeLittleCursors(self):
        """All non-active cursors try to keep up with the active one and try to reach the active tickindex.
        This is for going up and down between the tracks so you end up near your starting position.
        Technically it is to avoid a complete loop from the beginning each time you change the track to reach the old tickindex.
        This way the processing load is divided into incremental steps, executed at user action speed (much time for processing!)"""
        for x in self.cursors: # Try moving the current cursor again is redundant. But its a quick operation (just "pass" in the cursor class) and the cleaner code is worth the redundancy.
            x.goToTickIndex(self.activeCursor.tickindex, returnOnError = False)

    def left(self):
        """Cursor movement in sync. The active cursor steps left, gets the tickindex and all other cursors try to reach the tickindex."""
        returnvalue = self.activeCursor.left() #executes as well
        self.comeLittleCursors()
        return returnvalue #Return True or False to see if track beginning or not.

    def right(self):
        """Cursor movement in sync. he active cursor steps right, gets the tickindex and all other cursors try to reach the tickindex."""
        returnvalue = self.activeCursor.right() #executes as well
        self.comeLittleCursors()
        return returnvalue #Return True or False to see if track end or not.

    def measureLeft(self):
        """Move all cursors one measure of currently active cursor.ticks backwards.
        Going over a smaller timesig or different timesigs in tracks may result in valid but unexpected positions.
        Don't use for automated work except you know exactly what to expect (in a simple movement with only one timesig)"""
        oneMeasure = self.activeCursor.prevailingTimeSignature[-1].oneMeasureInTicks
        start = self.activeCursor.tickindex
        while self.activeCursor.tickindex > start - oneMeasure: #Go right until the current tick position is equal or bigger than the destination
            if not self.activeCursor.left(): #executes as well. First move the active cursor, later sync the others. So the cursor at least tries to go left. At the beginning of track it is executed twice during this loop.
                break # beginning of track
        self.comeLittleCursors()

    def measureRight(self):
        """see movement.measureLeft, only forward."""
        oneMeasure = self.activeCursor.prevailingTimeSignature[-1].oneMeasureInTicks
        start = self.activeCursor.tickindex
        while self.activeCursor.tickindex < start + oneMeasure: #Go right until the current tick position is equal or bigger than the destination
            if not self.right(): #executes as well. First move the active cursor, later sync the others. So the cursor at least tries to go right. At the end of staff it is executed twice during this loop.
                break # end of track
        self.comeLittleCursors()

    def head(self):
        for x in self.cursors:
            x.go_to_head()
            x.resetToDefault() #Reset all counters and gatherers to 0.

    def tail(self):
        while self.activeCursor.right(): #first the active cursor should go right.
            pass
        self.comeLittleCursors()

    def toTickIndex(self, ticks):
        """Go to a tickindex."""
        self.activeCursor.goToTickIndex(ticks, returnOnError = False)
        self.comeLittleCursors()

    def toIndex(self, index):
        """Go to the flat cursor index in the current track."""
        ret = self.activeCursor.goToIndex(index) #ret says if track start/end was encountered
        if ret:
            self.comeLittleCursors()
        else:
            return False

    #Tracks#
    def currentTrackIndex(self):
        """Return the index of the current track,
        which is the same as the active cursor"""
        return self.cursors.index(self.activeCursor)

    @property
    def currentTrack(self):
        """return the instance of the active track"""
        return self[self.currentTrackIndex()]

    def trackDown(self):
        """Change the active cursor to a higher index"""
        now = self.currentTrackIndex()
        nowTicks = self.activeCursor.tickindex
        try:
            self.cursors[now + 1]
            self.activeCursor = self.cursors[now + 1]
            return True
        except IndexError:
            return False  #Already the last Track
        #self.toTickIndex(nowTicks)  #TODO: never executed. But everything seems to work fine?!

    def trackUp(self):
        """Change the active cursor to a lower index"""
        now = self.currentTrackIndex()
        nowTicks = self.activeCursor.tickindex
        if now - 1 <= 0:
            self.activeCursor = self.cursors[0]
            return False
        else:
            self.activeCursor = self.cursors[now - 1]
            return True
        #self.toTickIndex(nowTicks) #TODO: never executed. But everything seems to work fine?!

    def lastTrack(self):
        """Change the active cursor to the highest index"""
        while self.trackDown():
            pass

    def firstTrack(self):
        """Change the active cursor to the lowest index"""
        while self.trackUp():
            pass

    def removeEmptyTracks(self):
        """Remove all tracks which have no duration items.
        Except the master track, which will be not touched"""
        emptyCursors = []
        for cursorEmptyTrack in self.cursors:
            if cursorEmptyTrack.nested_list.isEmpty() and not cursorEmptyTrack.nested_list.uniqueContainerName.lower() == "master":
                emptyCursors.append(cursorEmptyTrack)

        for emptyCursor in emptyCursors:
            self.cursors.remove(emptyCursor) #delete cursor
            self.container.remove(emptyCursor.nested_list) #delete track from movement container


    def parseAllTracks(self, command):
        """
        command() gets two parameters:
        the current item and the active cursor.
        The parsing stops if command returns the string "break".

        All other positive return values will be saved in a list
        which is returned in the end.
        It is a list of list, each track is one parent list.
        Return values are in the child list.

        The search is track based. Therefore the "break" keyword
        is of limited use if you want to search something until a
        certain amount of musical 'time' has passed because that search
        would be column based ("Parallel search in all tracks")

        The cursor stays at its position while parsing. This is not a
        "move cursor to found item" function.

        Best treat this function as "read only".

        Even a "simple" modifying task like "toggle Selection" could
        fail when dealing with links and containers. In this case an
        exception could be made if "Selection = True" is enforced.
        But even that has risks. If you want to select every second note
        and you only have links in the track it will constantly select
        and deselect the same note.
        The warning has been made… now I can use it myself to select :)

        In other words: you should not use "tone up" as command because
        linked items will be shifted up each time they are encountered
        The api commands like applyToSelection take care of these cases.

        Achtung! This does not work at all if the order is modified
        by the commands, especially if it is a duration change.
        In that case the old cursors will be loaded and the
        tickindex does not work anymore.
        If you want to modify the data use the api ApplyToSelection
        or a variant.
        """
        trackIndex = self.currentTrackIndex() #save the track position
        startTick =  self.activeCursor.tickindex #Save the tick index, nothing more. Work on the original cursors.
        gatherer = []
        #First prepare the structure that will be returned later. We need to build the empty one before gathering so it matches the track count.
        gatherer = [list() for tr in self.cursors] #we only need the number of cursors/tracks not the data
        try:
            for index, cursor in enumerate(self.cursors):
                self.activeCursor = cursor  #Set the active cursor so other functions (like the pitch module) can rely on it.
                self.activeCursor.resetToDefault()

                for obj in self.activeCursor.iterate():
                    ret = command(obj, self.activeCursor)
                    if ret == "break":
                        # raise an exception to break out of the nested loop
                        # this will not discard any gathered data (parentlist/childlist until now)
                        raise StopIteration()
                    elif ret:
                        gatherer[index].append(ret)
        except StopIteration:
            pass

        #Restore the old position. This should not have changed if the user/programmer read the docstring.
        self.activeCursor = self.cursors[trackIndex]
        self.toTickIndex(startTick)
        return gatherer

    def parseAllColumns(self, command):
        """
        See parseAllTracks, but column based:
        Track 2, Item 1 is considered 'earlier' than Track 1, last item.

        But it returns the same structure as parseAllTracks. Only useful
        if you use the 'break' keyword to stop the search at a certain
        point.

        Does not move the cursor

        Another difference is that the active cursor stays the same
        during parsing (contrary to parseAllTrack where it does not
        matter if you use the cursor parameter object or the active
        cursor). Here you have to use the cursor parameter object.
        """
        #TODO: replace tickindex with cursor Index.
        trackIndex = self.currentTrackIndex() #save the track position
        startTick =  self.activeCursor.tickindex #Save the tick index, nothing more. Work on the original cursors.
        #First prepare the structure that will be returned later. We need to build the empty one before gathering so it matches the track count.
        gatherer = [list() for tr in self.cursors] #we only need the number of cursors/tracks not the data
        for cursor in self.cursors:
            cursor.go_to_head()
            cursor.resetToDefault()

        #The real action. Two loops over all cursor in sequence. The first pass checks the data the second equalises unequal tick values so that we really follow a timeline.
        while True:
            currentTicks = [] #reset each time
            try:
                for index, cursor in enumerate(self.cursors):
                    ret = command(cursor.get_item(), cursor)
                    if ret == "break":
                        # raise an exception to break out of the nested loop. this will not discard any gathered data (parentlist/childlist until now)
                        raise StopIteration()
                    elif ret:
                        gatherer[index].append(ret)
                    if cursor.right(): #Only add the current tickindex if not at the end of track. Otherwise this will be always the minimum because other cursors advance further and we get an endless loop.
                        currentTicks.append(cursor.tickindex)
                if not currentTicks: #the search ends when the end of the track for every cursor was reached which means none of the cursors added its tickindex to this list.
                    raise StopIteration() #stop search. All cursors are at the end of their tracks.
                #Second pass. reset cursors that have a higher tickindex than the minimum
                for cursor in self.cursors:
                    if cursor.tickindex > min(currentTicks):
                        cursor.left()
            except StopIteration:
                #Restore the old position. This should not have changed if the user/programmer read the docstring.
                self.activeCursor = self.cursors[trackIndex]
                self.toTickIndex(startTick)
                return gatherer

    def analysisMovement(self):
        """returns a special copy of the current movement,
        based on unfoldMovement.

        Format is one list with tuplets. All tuplets have the same
        length/the same number of items.

        Details:
        The difference is that rests get replaced with the last pitch
        in each track.

        Also all tracks are made equal tick length by repeating the last
        pitch"""
        newMovement = self.unfoldMovement()

        length = []
        for cursor in newMovement.cursors:
            cursor.resetToDefault()
            cursor.go_to_head()
            track = cursor.nested_list
            track.container = [x for x in track.container if x.duration] #removes all non-music items. clefs and timesigs have no role here. All values like metrical position are already saved by unfoldMovement
            length.append(len(track.container))  #the list we get from unfoldMovement is already flat; Non-nested. This means we can use len(track.container) to get the actual number of items.
            last = Rest(384, "s")
            firstRealPitch = False #once this is true
            leadingRests = []
            for obj in cursor.iterate():
                obj.last = last
                firstPitch = obj.notelist[0].pitch
                lastPitch = last.notelist[0].pitch
                if firstPitch and not (firstPitch == "r" or firstPitch == "s"): #rests have "r" or "s" , multi measure has 0
                    last = obj
                    if not firstRealPitch: #we had only rests until here aka. this is the first chord.
                        for x in leadingRests: #all rests so far are considered the same as the current note.
                            x.notelist = obj.notelist
                        firstRealPitch = True
                        leadingRests = None
                elif lastPitch and not (lastPitch == "r" or lastPitch == "s") : #current obj is a rest, last one was not (or already overwritten)
                    obj.notelist = last.notelist #copy the last valid pitches to
                else: #current obj is a rest, last one was a rest as well. only happens with leading rests from the track start.
                    if leadingRests is None: #and not a full or empty list!
                        raise RuntimeError("Rests within the track were recognized as leading rests. This should not happen.")
                    leadingRests.append(obj) #save this for later


        maxItems = max(length)
        endRest = Rest(0, "s")
        #TODO
        return newMovement

    def convertToExportMovement(self):
        """This export movement keeps containers intact and is intended
        as a basis for Lilypond export.
        Returns a copy of the movement which is modified:
        find the mastertrack and seperate it from the other tracks.
        the masterCursor will be in movement.masterCursor and the
        mastertrack in movement.master.
        But the mastertrack and cursor get deleted from .cursors and
        .container"""

        newMovement = copy.deepcopy(self) #this copies all cursors including the prevailing sigs etc.

        #Don't process empty tracks. They are most likely just a place which was not edited yet but already created for later.
        newMovement.removeEmptyTracks()

        #The following is not needed for Lilypond. The lilypond code is just a string as far as we are concerned here. There is no way back to Laborejo objects.
        #for cursor in newMovement.cursors:
        #    cursor.go_to_head()
        #    cursor.resetToDefault()
        #    for obj in cursor.iterate():
        #        obj.flatCursorIndex = cursor.flatCursorIndex #for signals that go back from the unfolded Movement to the original movement we write down the unique position. This endObj.flatCursorIndex cannot be used in unfoldMovement because it is here to remember the original structure, without any repeats.
        #        obj.uniqueContainerName = cursor.nested_list.uniqueContainerName #for the same reasons as above remember the container name.

        for tr in newMovement.container:
            if tr.uniqueContainerName.lower() == "master":
                masterIndex = newMovement.container.index(tr)
                newMovement.masterCursor = newMovement.cursors[masterIndex]
                newMovement.cursors.pop(masterIndex)
                newMovement.master = tr
                newMovement.container.remove(tr)

                #flatten master list and filter it
                mastermembers = [KeySignature, Clef]
                newMovement.master.container = [item for item in newMovement.master.flatten() if type(item) not in mastermembers]
                break
        return newMovement

    def convertToUnfoldedExportMovement(self):
        """This is used for Playback, Analysis etc.
        not for Lilypond since Lilypond keeps containers intact.
        Returns a copy of the movement which is modified:
        find the mastertrack and seperate it from the other tracks.
        the masterCursor will be in movement.masterCursor and the
        mastertrack in movement.master.
        But the mastertrack and cursor get deleted from .cursors and
        .container"""

        newMovement = copy.deepcopy(self) #this copies all cursors including the prevailing sigs etc.

        #This does not need a master cursor exchange yet because we don't deal with pitch or timesigs. If that changes don't forget to add
        #originalActiveCursor = self.activeCursor
        #self.activeCursor = cursor  #Set the active cursor so other functions (like the pitch module) can rely on it.


        #Step Two: Flatten out containers.
        #This must be done early because when we add the flatcursor index to the items later they must be
        #and when the master track items are added eventually they must not be added to containers since they get duplicated then.

        for cursor in newMovement.cursors:
            cursor.go_to_head()
            cursor.resetToDefault()
            newFlatTrack = []
            tempRepeatPercents = {} #Percent Repeats for Containers
            for obj in cursor.iterate():
                typ = type(obj)
                if typ is not Start and typ is not End: #don't recursively copy the parent container in Start and End.
                    obj = copy.deepcopy(obj) #make the item unique
                obj.flatCursorIndex = cursor.flatCursorIndex #for signals that go back from the unfolded Movement to the original movement we write down the unique position. This endObj.flatCursorIndex cannot be used in unfoldMovement because it is here to remember the original structure, without any repeats.
                obj.uniqueContainerName = cursor.nested_list.uniqueContainerName #for the same reasons as above remember the container name.

                if typ is End: #container End
                    obj.durationSwitch = False #duration is the parent container duration here. remove before processing. This is not relevant for playback or analysis, only for the cursor and possible GUIs.
                    if obj.parentContainer.repeatPercent > 1:
                        if obj in tempRepeatPercents:
                            tempRepeatPercents[obj] -= 1
                        else: #first encounter
                            tempRepeatPercents[obj] = obj.parentContainer.repeatPercent

                        if tempRepeatPercents[obj] > 1: #even after substraction there is still one round left. Go back:
                            while True:
                                if cursor.left():
                                    it = cursor.get_item() #this is not the deep copy variant.
                                    if type(it) is Start and it.parentContainer is obj.parentContainer:
                                        break
                                else: #track beginning?!
                                    raise RuntimeError("This should never have happened. You have a Container End Item without a container Start item. Please contact the developers with your file!")
                        else: #the subtraction right now was from 1 to 0. This was already the last round. Remove obj from the temp storage if we encounter it again.
                            del tempRepeatPercents[obj]
                else: #any other item
                    if not typ is Start: #container start
                        newFlatTrack.append(obj)

            cursor.nested_list.container = newFlatTrack

        for tr in newMovement.container:
            if tr.uniqueContainerName.lower() == "master":
                masterIndex = newMovement.container.index(tr)
                newMovement.masterCursor = newMovement.cursors[masterIndex]
                newMovement.cursors.pop(masterIndex)
                newMovement.master = tr
                newMovement.container.remove(tr)
                break
        return newMovement

    def unfoldMovement(self):
        """Return an unfolded version of self. Unfolded means:
        -Merge repeats and jumps of the mastertrack to all tracks.
        -Unfold repeats in each track (which can result from merged in
        master repeats)
        -Unfold repeats in the mastertrack itself.
        -Unfold all substitutions, even the no-unfold ones.
        -Execute playback triggers
        -Give Grace Notes and Ornaments a musical meaning and shorten
        other notes.
        Lookaheads:
        -Last note of a slurred section is marked as not legato.

        This unfolded movement is the basis for playback export
        and analysis, but playback still needs to merge duration-objects
        from the master into main track.

        The unfolded movement is a mess to work with. It is just good
        for left-to-right exporting or checking.
        """

        newMovement = self.convertToUnfoldedExportMovement()
        originalActiveCursor = self.activeCursor

        #Step One: Delete all empty (no duration) tracks and solo.
        #This is not done by self.removeEmptyTracks() because we check for solo and mute here as well.

        #Create the solo mask.
        soloMode = any([tr.smfSolo for tr in newMovement])

        emptyCursors = []
        for cursorEmptyTrack in newMovement.cursors:
            if cursorEmptyTrack.nested_list.smfMute or (soloMode and not cursorEmptyTrack.nested_list.smfSolo) or cursorEmptyTrack.nested_list.isEmpty():
                emptyCursors.append(cursorEmptyTrack)

        for emptyCursor in emptyCursors:
            newMovement.cursors.remove(emptyCursor) #delete cursor
            newMovement.container.remove(emptyCursor.nested_list) #delete track from movement container

        #Now only real tracks are in here.

        if newMovement.masterCursor:
            #from now on work with the seperated mastertrack and all other tracks as a group.
            self.activeCursor = newMovement.masterCursor  #Set the active cursor so other functions (like the pitch module) can rely on it.

            masterList = [] # a list of tuplets: (tickPos, object)

            newMovement.masterCursor.resetToDefault()
            newMovement.masterCursor.go_to_head()
            for obj in newMovement.masterCursor.iterate():
                if obj.duration == 0 and not type(obj) in [Start, End, Appending, Clef, KeySignature, Clef]:
                    masterList.append((newMovement.masterCursor.tickindex, obj))

            #Insert the content of the mastertrack to the other tracks so that the mastertrack can be ignored from now onw
            for cursor in newMovement.cursors:
                self.activeCursor = cursor
                cursor.resetToDefault()
                cursor.go_to_head()
                lastPos = 0
                for pos, obj in masterList:
                    if not pos == cursor.tickindex:
                        if not cursor.goToTickIndex(pos, returnOnError = False): #Move to the right. Don't go back if the tickindex at the end is not the same as desired.
                            if cursor.tickindex >= pos:
                                raise ValueError("Target tick position", pos , "is beyond this tracks final tickindex", cursor.tickindex , " but the current tickindex is >= than the target position. This is logically not possible")
                            diff = pos - cursor.tickindex
                            cursor.insert(Rest(diff, "s"))

                        #If the tickindex was inside a multi measure rest or long note we have to split it, which is no problem for rests.
                        #In this case we already jumped over the tickindex and have to go one step back to get the long duration item.
                        elif cursor.tickindex > pos:
                            cursor.left()
                            it = cursor.get_item()
                            if type(it) is MultiMeasureRest:
                                cursor.delete()
                                #Since there is no way there is any change whatever inside a MMRest we only need to get the cursor prevailing data only once.
                                timeSigNominator = cursor.prevailingTimeSignature[-1].nominator #example: x/384. Upper number
                                timeSigDenominator = cursor.prevailingTimeSignature[-1].denominator #example: 4/x. Lower number
                                #Now that the MMRest is deleted insert small rests instead. A rule of thumb is a
                                for measurenumber in range(it.measures): #For each measure do(counts from 0)
                                    for metricalPos in range(timeSigNominator): #do this n times per measures. for example 3 times in a 3/4 timesig.
                                        cursor.insert(Rest(timeSigDenominator, "s")) #insert a rest with the lower timesig number. 384 in a 3/4 timesig.

                        #Attempt a second time to go to the correct tickindex.
                        if cursor.tickindex != pos:
                            raise RuntimeError("The  item", obj, obj.exportLilypond() , "from the master track couldn't be inserted into track", cursor.nested_list.uniqueContainerName, "since there is an item", cursor.get_item(), cursor.get_item().exportLilypond(), " spanning over tickindex", cursor.tickindex, ". Please correct manually!")

                    cursor.insert(obj) #insert goes right automatically.


            cursorsPlusMasterCursor = newMovement.cursors
            cursorsPlusMasterCursor.append(newMovement.masterCursor)

        else: #no master track.
            cursorsPlusMasterCursor = newMovement.cursors

        #Repeats
        #Only strictly valid repeats are considered. Every repeat close needs a repeat open. There are no implicit traditions here.

        for cursor in cursorsPlusMasterCursor:
            ##### First Pass ####
            #Copy linked items to unqiue items.
            #Gather Information about alternating ends [1… :|| [2…
            #Lookahead like "not legato at the end of slur" markers.
            self.activeCursor = cursor
            cursor.go_to_head()
            cursor.resetToDefault()
            firstPassContainer = [] #For first pass: a new, flat, data structure that will replace the track container in newMovement
            lastPassedEnding = None
            alternateEndingGroupCounter = 0 #a counter
            alternateEndingsDict = {} # A dict of dicts. Key is the group index above, value is another dict with: key = number of ending, value = flat Cursor Index

            for endObj in cursor.iterate():
                #Create copies of new linked items, container items etc. Create copies for all items. This is much more robust.
                #Achtung: This only works because we are adding "endObj" at the end to a new list. This gets not written back to the original track list.
                endObj = copy.deepcopy(endObj)
                endTyp = type(endObj)
                plainObjectAndNotSubstituted = True # We assume a standard item, not a chord with substitution.

                if endTyp is Chord and endObj.substitution:
                    plainObjectAndNotSubstituted = False #switch so that the parent item itself will not be added.
                    firstPassContainer += endObj.unfoldSubstitution(cursor.prevailingKeySignature[-1]) #Add the returned list of substitution items instead of the parent item

                elif endTyp is AlternateEnd:
                    lastPassedEnding = endObj

                    if 1 in endObj.endings:
                        alternateEndingGroupCounter += 1
                    endObj.alternateEndingGroup = alternateEndingGroupCounter
                    if not alternateEndingGroupCounter in alternateEndingsDict:
                        alternateEndingsDict[alternateEndingGroupCounter] = {}
                    for end in endObj.endings:
                        alternateEndingsDict[alternateEndingGroupCounter][end] = cursor.flatCursorIndex

                elif (endTyp is RepeatClose or endTyp is RepeatCloseOpen):
                    if lastPassedEnding:# and len(lastPassedEnding.endings) >= 2: #add additional repeats to the next repeat close if this is ending [1,2,3 [4...
                        endObj.repeat = len(lastPassedEnding.endings)   #force the repeat count to be exactly the sum of ending numbers.
                    lastPassedEnding = None

                #Finally add the item if it wasn't substituted. Else it was already added a few lines above.
                if plainObjectAndNotSubstituted:
                    firstPassContainer.append(endObj)


            #Get the last ending(number) for each group and save it as [-1]
            for groupIndex, positionDict in alternateEndingsDict.items():
                flatPos = sorted(positionDict.items())[-1][1]  #from the last tuple in the sorted, converted list get the second, which is the flat cursor index of the last AlternateEnding of this group
                alternateEndingsDict[groupIndex][-1] = flatPos
            cursor.nested_list.container = firstPassContainer #replace original, nested, data list with flat, unfolded new one.
            #We now have a dict with all groups, endings and cursor positions. If we encounter an AlternateEnd object after a repeat in the main loop below we can get its group and jump to the next AlternateEnd in line


            #####Second Pass: Main Pass####
            #reset again and start unfolding
            cursor.go_to_head()
            cursor.resetToDefault()
            activated = True #Repeats are activated by default. Can be deactivated by triggers.
            openStack = [] # a stack with plain-track-indices to indicate the position of the last repeat open |:
            closeDict = {} # a dict for repeat close. key is the plain-track-index, value is the number of repeats left. The initial value comes from the repeat close object.
            segnoPos = None
            gotosegno = True #Once segno is done this is switched on so the next time D.S. is ignored.
            gotocapo = True
            breakAt = None
            alternativeEndingsGroupsStatus = {} #Key = group number, Value = Ending number status. Keeps track what of what was the last alternative ending so that nonlinear structures like [1,3 [2,4  work.
            newContainer = [] #For second pass: a new, flat, data structure that will replace the track container in newMovement
            lastParsedChord = None
            subitoDynamic = None #sfz, sfp etc.

            def goToIndexAndSignatureInsert(cursor, position):
                """Returns a bunch of signatuers that must be added to
                the currently created container"""
                cursor.goToIndex(position)
                ret = []
                for signature in cursor.getState()["prevailingSignatures"]: #Reset all signatures to the current state by inserting redundant items. They will be exportPlayback later.
                    ret.append(signature)
                return ret

            for obj in cursor.iterate():
                typ = type(obj)
                if typ is Chord:
                    #Next iteration. Save the current object
                    #previousObjWithDuration = obj
                    if subitoDynamic:
                        obj.subitoDynamicKeyword = subitoDynamic
                        subitoDynamic = None

                elif typ is Rest:
                    pass #performance. Skip the whole chain.

                #elif typ is Ornament:

                elif typ in [SlurOff, PhrasingSlurOff]:
                    #We need to mark the last encountered chord as "not legato" which is a playback exception.
                    if lastParsedChord:
                        lastParsedChord._lookaheadMarkers["slurEnd"] = True
                    else:
                        raise RuntimeError("There was a slur close without even a chord before. This makes no sense.")


                elif typ is AlternateEnd:
                    if activated:
                        if not obj.alternateEndingGroup in alternativeEndingsGroupsStatus :  #first encounter of a group, typically. ending number one
                            alternativeEndingsGroupsStatus[obj.alternateEndingGroup] = 0 #initialize
                        alternativeEndingsGroupsStatus[obj.alternateEndingGroup] += 1
                        nextPos = alternativeEndingsGroupsStatus[obj.alternateEndingGroup]
                        newContainer += goToIndexAndSignatureInsert(cursor, alternateEndingsDict[obj.alternateEndingGroup][nextPos])

                    else: #not activated and we encounter an ending. We need to jump to the last ending now.
                        lastPos = alternateEndingsDict[obj.alternateEndingGroup][-1] #-1 is the key in the subdict for the position of the last Ending.
                        newContainer += goToIndexAndSignatureInsert(cursor, lastPos)

                elif typ is SubitoDynamicSignature:
                    subitoDynamic = obj.expression

                elif typ is Fine and breakAt is Fine:
                    break
                elif typ is Segno:
                    segno = cursor.flatCursorIndex
                elif typ is GotoSegno:
                    if gotosegno and (segno or segno == 0):
                        newContainer += goToIndexAndSignatureInsert(cursor, segno)
                        activated = not (obj.disableRepeat)
                        gotosegno = False
                        breakAt = obj.until
                        obj = cursor.get_item() #get the new item when for appending to newContainer a few lines further #TODO: Grammar
                        typ = type(obj)
                        closeDict = {} #reset all repeats. Jumping over repeats is done via the gotocapo switch, so this is safe and needed for alternate endings.
                        for gr, value in alternativeEndingsGroupsStatus.items(): #re-initiliaze/reset all alternate ending status
                            alternativeEndingsGroupsStatus[gr] = 0
                elif typ is GotoCapo:
                    if gotocapo:
                        activated = not (obj.disableRepeat)
                        gotocapo = False
                        breakAt = obj.until
                        newContainer += goToIndexAndSignatureInsert(cursor, 0) #the problem now is that the first item needs to get processed again but it will continue with the next item. It will be added later, but not processed.
                        obj = cursor.get_item() #get the new item when for appending to newContainer a few lines further #TODO: Grammar
                        typ = type(obj)
                        closeDict = {} #reset all repeats. Jumping over repeats is done via the gotocapo switch, so this is safe and needed for alternate endings.
                        for gr, value in alternativeEndingsGroupsStatus.items(): #re-initiliaze/reset all alternate ending status
                            alternativeEndingsGroupsStatus[gr] = 0
                elif typ is GotoCoda:
                    if breakAt is Coda:
                        while True:
                            if cursor.right():
                                if type(cursor.get_item()) is Coda:
                                    obj = cursor.get_item() #get the new item when for appending to newContainer a few lines further
                                    break
                            else: #track end
                                warnings.warn("Instructed to go to Coda. But no Coda found. The coda must be right of the GotoCoda item")
                                break
                        gotocoda = False

                if activated: #We are not in a Da Capo iteration with deactivated repeats
                    if typ is RepeatOpen:
                        openStack.append(cursor.flatCursorIndex)
                    elif typ is RepeatClose or typ is RepeatCloseOpen:
                        if cursor.flatCursorIndex in closeDict: #we were here already
                            if closeDict[cursor.flatCursorIndex]: # number of repeats > 0
                                closeDict[cursor.flatCursorIndex] -= 1
                                newContainer += goToIndexAndSignatureInsert(cursor, openStack[-1])
                            else: #We were here already and there are no repeats left
                                openStack.pop()
                                del closeDict[cursor.flatCursorIndex]
                                if typ is RepeatCloseOpen: #We got behind the Repeat Close. We consider this item as new RepeatOpen and move on.
                                    openStack.append(cursor.flatCursorIndex)
                        else: #new encounter of a repeat close
                            if openStack: #we are in a valid repeat open section
                                closeDict[cursor.flatCursorIndex] = obj.repeat - 1 #add obj to the close dict as position index and number of repeats, already decreased -1 because we are here the first time right now.
                                newContainer += goToIndexAndSignatureInsert(cursor, openStack[-1])
                            else:
                                warnings.warn("Repeat close without repeat open. No implicit 'Repeat Open' in Laborejo! Consider a hidden repeat open.")


                #Playback Triggers.
                obj.playbackCounter += 1
                if obj.triggerString is None or obj.playbackTrigger():
                    if not (isinstance(obj, SpecialBarline) or isinstance(obj, GotoSegno) or isinstance(obj, GotoCapo) or isinstance(obj, GotoCoda)) : #TODO: this produces clearer lists but could be a small performance issue
                        newContainer.append(obj)

                else: #The playback trigger returned false. Insert an empty item of the same duration.
                    emptyItem = Rest(obj._base, "s") #a replacement instance
                    emptyItem._base = obj._base
                    emptyItem._tuplets = obj._tuplets
                    emptyItem._dots = obj._dots
                    emptyItem._scaleFactorNumerator = obj._scaleFactorNumerator
                    emptyItem._scaleFactorDenominator = obj._scaleFactorDenominator
                    newContainer.append(emptyItem)

                lastParsedChord = obj

            cursor.nested_list.container = newContainer #replace original, nested, data list with flat, unfolded new one.
        self.activeCursor = originalActiveCursor


        #TODO: Change original active cursor in all functions to a parameter.
        #TODO: So we can work with the current cursor.

        return newMovement

    def addTrack(self):
        """Add a Track as last Track"""
        tr = Track()
        self.append(tr) #Add the new Track to the movement container. It has a unique name on init which can be replaced later by the user.
        newCur = cursor.Cursor(track = self[-1], parentMovement = self) #add last track to be controlled by this cursor and give reference to self, the parentMovement.
        self.cursors.append(newCur) #Create a new cursor for the new Track
        if len(self.cursors) == 1:  #for the first cursor: make it active
             self.activeCursor = self.cursors[0]
        self.lastTrack()
        self.comeLittleCursors()
        return tr

    def deleteTrack(self):
        """Delete one track. New active track is the new track on the
        same position."""
        current = self.currentTrackIndex()
        del self.cursors[current]  #del removes by index
        del self[current] #remove the track by its index.

        if self: #Is still a Track left or was that the last one?
            try:
                self.cursors[current]  #is there still a track on the old position.
                self.activeCursor = self.cursors[current] #if yes it is fine. make the new cursor/track on this position the active one.
            except:  #if not we deleted the last track.
                self.activeCursor = self.cursors[-1]  #make the new last cursor the active one.
        else:
            self.addTrack() #Create a new empty Track instead.
            self.activeCursor = self.cursors[0] #and reset the cursor

    def swapTrackWithAbove(self):
        """Move the complete Track up which results in the upper Track moving down"""
        now = self.currentTrackIndex()
        if now == 0: #Top Track cannot move up
            return False #no swap done.
        else:
            self.container.insert(now - 1 , self.container.pop(now)) #swap the track
            self.cursors.insert(now - 1 , self.cursors.pop(now)) #swap the cursor
            self.activeCursor = self.cursors[now - 1] #new active cursor is the old one.
            return True

    def swapTrackWithBelow(self):
        """Move the complete Track down which results in the lower Track moving up"""
        now = self.currentTrackIndex()
        if now == len(self.cursors) - 1: #Bottom Track cannot move down
            return False #no swap done.
        else:
            self.container.insert(now + 1 , self.container.pop(now)) #swap the track
            self.cursors.insert(now + 1 , self.cursors.pop(now)) #swap the cursor
            self.activeCursor = self.cursors[now + 1] #new active cursor is the old one.
            return True

    def createContainerFromSelection(self):
        """Take all selected items, create a new container from them.
        Container are single-track. If this is a multi selection the
        items will be merged as single track.

        The new container will only have copies of the original items."""
        tempList = []

        start = (self.activeCursor, self.activeCursor.tickindex) #Save the cursor position for later
        for cursor in self.cursors:
            track = [i for i in cursor.iterate() if i.selected] #create a subset of the track with all selected items.
            if track: #the step above also creates empty lists for tracks with no selection.
                tempList.extend(copy.deepcopy(track)) #extend the single list with a copy of these instances

        for y in tempList:
            y.instanceCount = 0
            y.deselect(False) #set status of the item to 'not selected' but don't try to delete from the selectedItems list nor decrease selection count. The copied item is of course not in there.

        self.head() #return to the beginning, resets all cursors.
        self.activeCursor = start[0]
        self.toTickIndex(start[1])
        if tempList: #empty?
            new = Container()
            new.container = new.container[:1] + tempList + new.container[-1:]
            return new
        else:
            return None

    def generateSubtext(self):
        """Substitute line breaks and paragraphs."""
        escaped = self.subtext.replace("<br>", " }\n \\wordwrap-lines { ")
        escaped = escaped.split("\n\n")
        if self.subtext:
            return "\n".join(["\\wordwrap-lines { " + line + " } \\vspace #0.5" for line in escaped])
        else:
            return "\\wordwrap-lines { }"

    def generateLilypondContainer(self):
        """Generate the container definitions for Lilypond export"""
        containerDefinitionsString = "\n".join([key + " = { " + value + " } " for key, value in self.nonTrackContainer.items()])
        self.nonTrackContainer = {}
        if self.master:
            masterVoice = self.master.generateLilypondVoice()
            masterTrack = "masterTrack = { " + self.master.exportLilypond() + " }\n"
        else:
            masterTrack = masterVoice = ""
        return masterVoice + masterTrack + containerDefinitionsString

    def generateLilypondVoice(self):
        """Lilypond Voice definitions block"""
        #Reset special Lilypond Export Data for each Movement. These are not save values.
        #session.workspace.lastDurationLyString = ""
        session.workspace.lilypondExportPrevailingTimeSignature = session.workspace.movement.activeCursor.prevailingTimeSignature[0] #Holds one TimeSignature at a time, defaults to Lilyponds 4/4. Is overwritten and used during Lilypond export and gets accessed and changed by Multi Measure Rests and other methods and functions. Similar to the cursor.prevailingTimeSignature but does not go left.
        session.workspace.waitForNextChordPst = [] # A queue which will be attached as postfix directives to the next chord in line. Enables dynamic changes to be standalone items but still appear as attached lilypond.
        session.workspace.waitForNextChordPre = [] # A queue which will be attached as postfix directives to the next chord in line. Enables dynamic changes to be standalone items but still appear as attached lilypond.
        return "".join([track.generateLilypondVoice() for track in self.container]) + "\n"

    def generateLilypond(self):
        """Structure: Staff Groups (through directives), staffs, voices.
        Lilypond has only one voice level so we don't need nested voices
        level checking like tuplets (except in-staff micro voices which
        have another syntax here and in Lilypond)
        The internal "Tracks" are arranged as Lilypond voices and staffs."""
        def structGen(x):
            """Build a structure line. Decide if opening or closing <<>>
            are needed to close a staff which may have multiple voices"""
            currentStaffIndex = self.container.index(x)
            #The prefix << is in staff itself so a staffgroup can be added here.
            try: #next staff present at all and still a voice? If not close staff.
                self[currentStaffIndex + 1] and not self[currentStaffIndex + 1].hideLilypond
                if self[currentStaffIndex + 1].mergeWithUpper:
                    post = " " #next is still a visible voice. Do nothing fancy
                else:
                    post = ">>\n" #next is present but not a merged voice. Close current staff.
            except IndexError:
                    post = ">>\n"  #Already the last staff

            return x.exportLilypond() + post

        def lyrics(x):
            """x = track
            Build a lyrics line.
            Wants a track as parameter"""
            if x.hideLilypond or (not x.lyrics):
                return ""
            else:
                return "".join(["\\context Lyrics = \"" + x.uniqueContainerName  + "Lyrics" + helper.num2word(verseCounter) + "\" \\lyricsto \"" + x.uniqueContainerName + "\" \\" + x.uniqueContainerName + "Lyrics" + helper.num2word(verseCounter) + "\n" for verseCounter in range(x.verseCountCache)])

        return  "".join([structGen(x) for x in self if (not x.hideLilypond)]) + "\n" + "".join([lyrics(x) for x in self])

    def exportPlayback(self, unfoldedMovement = None, jackMode = False):
        """The original cursor positions are backuped and restored"""
        if not unfoldedMovement:
            unfoldedMovement = self.unfoldMovement()
        originalActiveCursor = self.activeCursor #Save the cursor position
        smfStructure = smf.SMF(ppqn=384)
        trackNum = 0
        #Work on the unfolded copy, not on the original cursors
        #TODO: We still use the real active cursor. Unless all functions don't rely on this state anymore we can't change it.
        self.activeCursor =  unfoldedMovement.activeCursor

        for track in unfoldedMovement: #this does not go through the tracks. The tracks just provide a number. The cursors point to the real data tracks.
            smfStructure.add_track()
            self.activeCursor = unfoldedMovement.cursors[trackNum]
            self.activeCursor.resetToDefault()

            if trackNum == 0: #only for the first track.
                if "copyright" in session.workspace.header.container:
                    smfStructure.tracks[0].add_event(smf.Event([0xFF, 0x02, len(session.workspace.header.container["copyright"].value)] + [ord(c) for c in session.workspace.header.container["copyright"].value]), pulses = 0)  #Copyright.

            #Decide if standard SMF or the Jack Set of parameters
            channel = track.jackChannel if jackMode else track.smfChannel
            volume = track.jackVolume if jackMode else track.smfVolume
            pan = track.jackPan if jackMode else track.smfPan
            bank = track.jackBank if jackMode else track.smfBank
            patch = track.jackPatch if jackMode else track.smfPatch
            transpose = track.jackTransposition if jackMode else track.smfTransposition

            smfStructure.tracks[trackNum].add_event(smf.Event([0xFF, 0x03, len(track.uniqueContainerName)] + [ord(c) for c in track.uniqueContainerName]), pulses = 0)  #Track Name
            smfStructure.tracks[trackNum].add_event(smf.Event([0xFF, 0x04, len(track.longInstrumentName)] + [ord(c) for c in track.longInstrumentName]), pulses = 0)  #Initial Instrument Name. Can be changed by the object event.
            initTempoSig = self.activeCursor.prevailingTempoSignature[0].generatePlayback(0,0,0)[0][0]
            smfStructure.tracks[trackNum].add_event(smf.Event(initTempoSig), pulses = 0) #Initial Tempo. Global event, not for any channel. Fake 0,0,0 parameters
            #Lyrics, 0x04, need an de-lilypondized version. Frescobaldi can do that.
            #smfStructure.tracks[trackNum].add_event(smf.Event([0xFF, 0x05, len(track.longInstrumentName)] + [ord(c) for c in track.longInstrumentName]), pulses = 0)  #Initial Instrument Name. Can be changed by the object event.

            if jackMode and channel[1] > channel[0]: #a real range
                for i in range(channel[0], channel[1]+1):
                    smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + i, 0x07, volume]), pulses = 0,) #Initial Volume for this channel. CC 7
                    smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + i, 0x0A, pan]), pulses = 0,) #Initial Pan (Left/Right). for this channel. CC 10
                    smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + i, 0x00, bank]), pulses = 0,) #Initial Bank for this channel. CC0
                    smfStructure.tracks[trackNum].add_event(smf.Event([0xC0 + i, patch]), pulses = 0,) #Initial Patch/Instrument

            else:
                smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + channel[0], 0x07, volume]), pulses = 0,) #Initial Volume for this channel. CC 7
                smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + channel[0], 0x0A, pan]), pulses = 0,) #Initial Pan (Left/Right). for this channel. CC 10
                smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + channel[0], 0x00, bank]), pulses = 0,) #Initial Bank for this channel. CC0
                smfStructure.tracks[trackNum].add_event(smf.Event([0xC0 + channel[0], patch]), pulses = 0,) #Initial Patch/Instrument

            #track.exportPlayback(smfStructure.tracks[trackNum], ticks = 0, channel = channel, transpose = track.smfTransposition)  #each track starts at 0 smf-ppqn and has a basechannel
            workingChannel = [channel[0], channel[1], channel[0]] #during object export we keep the original channel and also a working channel for note ond and off permanent/standalone ChannelChange items
            for obj in self.activeCursor.iterate():
                smfTrack, workingChannel, transpose, jackMode = obj.exportPlayback(smfStructure.tracks[trackNum], channel = workingChannel, transpose = transpose, jackMode = jackMode)

            #The cursor now is on the beginning of the last tick, which is appending. So all sounding events are over.
            #Some midi players shut off playback on the last note off or at the EndOfTrack object and don't wait for any reverbs or sounding notes.
            #We add a Sustain Pedal Off after the last event so we get a small buffer time. This also covers forgotton sustain pedals.
            if jackMode and channel[1] > channel[0]: #a real range
                for i in range(channel[0], channel[1]+1):
                    smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + i, 0x40, 0]), pulses=self.activeCursor.tickindex + 1536) #CC 64, sustain pedal off.
            else:
                smfStructure.tracks[trackNum].add_event(smf.Event([0xB0 + channel[0], 0x40, 0]), pulses=self.activeCursor.tickindex + 1536) #CC 64, sustain pedal off.

            trackNum += 1


        self.activeCursor = originalActiveCursor #Restore original active cursor including its position.
        if trackNum == 0: #all tracks are muted
            smfStructure.add_track() #add one empty pseudo track.
        return smfStructure

    def exportPlaybackTracks(self, unfoldedMovement = None, jackMode=False):
        """Like self.exportPlayback but returns a list of smf structures
        Each track is one file."""
        if not unfoldedMovement:
            unfoldedMovement = self.unfoldMovement()
        originalActiveCursor = self.activeCursor #Save the cursor position
        #Work on the unfolded copy, not on the original cursors
        #TODO: We still use the real active cursor. Unless all functions don't rely on this state anymore we can't change it.
        self.activeCursor =  unfoldedMovement.activeCursor

        trackNum = 0 #This is not the same as in exportPlayback(). We only use it for the internal cursor numbers since all midi tracks are seperate files anyway.
        returnedStructures = []

        for track in unfoldedMovement: #this does not go through the tracks. The tracks just provide a number. The cursors point to the real data tracks.
            smfStructure = smf.SMF(ppqn=384)
            smfStructure.add_track()
            if "copyright" in session.workspace.header.container:
                smfStructure.tracks[0].add_event(smf.Event([0xFF, 0x02, len(session.workspace.header.container["copyright"].value)] + [ord(c) for c in session.workspace.header.container["copyright"].value]), pulses = 0)  #Copyright.

            #Decide if standard SMF or the Jack Set of parameters
            channel = track.jackChannel if jackMode else track.smfChannel
            volume = track.jackVolume if jackMode else track.smfVolume
            pan = track.jackPan if jackMode else track.smfPan
            bank = track.jackBank if jackMode else track.smfBank
            patch = track.jackPatch if jackMode else track.smfPatch
            transpose = track.jackTransposition if jackMode else track.smfTransposition

            smfStructure.tracks[0].add_event(smf.Event([0xFF, 0x04, len(track.longInstrumentName)] + [ord(c) for c in track.longInstrumentName]), pulses = 0)  #Initial Instrument Name. Can be changed by the object event.
            initTempoSig = self.activeCursor.prevailingTempoSignature[0].generatePlayback(0,0,0)[0][0]
            smfStructure.tracks[0].add_event(smf.Event(initTempoSig), pulses = 0,) #Initial Tempo. Global event, not for any channel. Fake 0,0,0 parameters

            if jackMode and channel[1] > channel[0]: #a real range
                for i in range(channel[0], channel[1]+1):
                    smfStructure.tracks[0].add_event(smf.Event([0xB0 + i, 0x07, volume]), pulses = 0,) #Initial Volume for this channel. CC 7
                    smfStructure.tracks[0].add_event(smf.Event([0xB0 + i, 0x0A, pan]), pulses = 0,) #Initial Pan (Left/Right). for this channel. CC 10
                    smfStructure.tracks[0].add_event(smf.Event([0xB0 + i, 0x00, bank]), pulses = 0,) #Initial Bank for this channel. CC0
                    smfStructure.tracks[0].add_event(smf.Event([0xC0 + i, patch]), pulses = 0,) #Initial Patch/Instrument
            else:
                smfStructure.tracks[0].add_event(smf.Event([0xB0 + channel[0], 0x07, volume]), pulses = 0,) #Initial Volume for this channel. CC 7
                smfStructure.tracks[0].add_event(smf.Event([0xB0 + channel[0], 0x0A, pan]), pulses = 0,) #Initial Pan (Left/Right). for this channel. CC 10
                smfStructure.tracks[0].add_event(smf.Event([0xB0 + channel[0], 0x00, bank]), pulses = 0,) #Initial Bank for this channel. CC0
                smfStructure.tracks[0].add_event(smf.Event([0xC0 + channel[0], patch]), pulses = 0,) #Initial Patch/Instrument

            self.activeCursor = unfoldedMovement.cursors[trackNum]
            self.activeCursor.resetToDefault()
            workingChannel = [channel[0], channel[1], channel[0]] #during object export we keep the original channel and also a working channel for note ond and off permanent/standalone ChannelChange items
            for obj in self.activeCursor.iterate():
                smfTrack, workingChannel, transpose, jackMode = obj.exportPlayback(smfStructure.tracks[0], channel = workingChannel, transpose = transpose, jackMode = jackMode)

            #The cursor now is on the beginning of the last tick, which is appending. So all sounding events are over.
            #Some midi players shut off playback on the last note off or at the EndOfTrack object and don't wait for any reverbs or sounding notes.
            #We add a Sustain Pedal Off after the last event so we get a small buffer time. This also covers forgotton sustain pedals.
            if jackMode and channel[1] > channel[0]: #a real range
                for i in range(channel[0], channel[1]+1):
                    smfStructure.tracks[0].add_event(smf.Event([0xB0 + i, 0x40, 0]), pulses=self.activeCursor.tickindex + 1536) #CC 64, sustain pedal off.
            else:
                smfStructure.tracks[0].add_event(smf.Event([0xB0 + channel[0], 0x40, 0]), pulses=self.activeCursor.tickindex + 1536) #CC 64, sustain pedal off.

            #Append a tuplet to the list-of-return
            #0: smf Structure
            #1: unique track name
            #2: jack port
            #3+4: optional jack audio return ports left and right. only used for Ardour3 export or other exporters with Audio recording functionality.
            returnedStructures.append((smfStructure, track.uniqueContainerName, track.jackPort, int(self.activeCursor.tickindex), track.jackAudioReturnPortLeft, track.jackAudioReturnPortRight))

            #Eventhough everything is track 0 we still increase the counter for the internal cursor/track references. (self.activeCursor = unfoldedMovement.cursors[trackNum])
            trackNum += 1

        self.activeCursor = originalActiveCursor #Restore original active cursor including its position.
        return returnedStructures

    def generateSave(self):
        childDict = {}
        childDict["subtext"] = self.subtext
        childDict["lyFontSize"] = self.lyFontSize
        childDict["defaultPerformanceSignature"] = self.defaultPerformanceSignature.exportSave()
        subsaveDict = {}
        for key, value in self.substitutions.items():
            subsaveDict[key] = value.exportSave()
        childDict["substitutions"]  = subsaveDict
        #childDict["master"] = self.master.exportSave()
        return childDict

class ExportContainer(object):
    """This is one file. Open, New all results in one ExportContainer.
    Each var exports as a %$$PLACEHOLDER$$ in Lilypond export."""
    def __init__(self):
        #export data
        self.lilypondDefinitions = LilypondDefinitions()
        self.emptyHeader = Header() #we need an always empty, never editable header for lilypond export. When it is decided if this is a standalone piece or movement and which header position to use, in the \score or outside.
        self.header = Header()
        self.uniqueContainerNames = [] #all unique names for containers, tracks and movement. they must be unique for save and load as well as exporting to Lilypond.
        self.movement = None #The current movement.
        #various data for lilypond export
        self.waitForNextChordPst = [] #a temp value which gets filled during Ly Export. Needed because some Laborejo standalone items are postfix strings in Lilypond
        self.waitForNextChordPre = [] #a temp value which gets filled during Ly Export. Needed because some Laborejo standalone items are prefix strings in Lilypond
        #self.lastDurationLyString = None #For cleaner Lilypond export. The duration is only needed when it changes, not all the time. class Chord uses this during Ly Export to remember the last duration.
        self.lytemplate = "basic.ly" #For the whole file. No per-movement-templates. Can be changed by the user.
        self.lastKnownFilename = "" # Temporary path+filename storage for the whole File. Is not saved or used for anything critical. Just in lilypond export to write the filename in the ly file, to a comment line. Exception: SaveAll uses this to check if a file is worth saving. saveAll is a mean function anyway...
        self.addMovement()
        self.parentSession = None #once loaded or implemented via session.new this becomes the current session. not saved.

    def addMovement(self):
        mov = Movement()
        self.movement = mov #make active
        #self.allMovements.append(mov) #store

    def exportSave(self):
        dictionary = {}
        dictionary["file"] = "laborejo" #This makes .lbjs a Laborejo file.
        dictionary["lilypondDefinitions"] = self.lilypondDefinitions.exportSave()
        dictionary["header"] = self.header.exportSave()
        dictionary["movement"] = self.movement.exportSave()
        dictionary["lytemplate"] = self.lytemplate
        dictionary["uniqueContainerNames"] = self.uniqueContainerNames
        return dictionary

class SessionClipboard(object):
    """This is outside of any music structure so that different files
    and movements during the same session can share data.

    This is not saved in any file. If you close the program it is gone"""
    def __init__(self):
        self.lastMovement = None #on copy save the last movement. paste in api will check if it is pasting in the same movement. If not it will remove links.
        self.clipboard = {"session":[],} #value is a list. temp storage if a user script wants to do copy and paste action without touching the actual clipoard
        self.linkboard = {"session":[],} #value is a list. temp storage if a user script wants to do copy and paste action without touching the actual clipoard

    def copySelectedItemsToInternalClipboard(self, cliplink): #cliplink is a hash for the two dicts
        """Take all selected items, create a new list from them.
        No positional information is saved except in how many staffs was
        selected"""
        self.lastMovement = session.workspace.movement
        tmpclipboard = []
        tmplinkboard = []

        start = (session.workspace.movement.activeCursor, session.workspace.movement.activeCursor.flatCursorIndex) #Save the active cursor and its position for later
        for cursor in session.workspace.movement.cursors:
            track = [i for i in cursor.iterate() if i.selected] #create a subset of the track with all selected items.
            if track: #the step above also creates empty lists for tracks with no selection.
                tmpclipboard.append(copy.deepcopy(track)) #store a copy of these instances each for paste(normal). We don't know what is going to happen right now so we need both.
                tmplinkboard.append(track) #store the exact instances for paste as link. We don't know what is going to happen right now so we need both.

        for x in tmpclipboard: #each track in the clipboard
            for y in x: #each item in a track
                y.instanceCount = 0
                #y.newName() #is done during api.paste because we can paste an item more than once. Just generating a new name on copy would result in pasting the same thing over and over again.
                y.deselect(False) #set status of the clipboard-item to 'not selected' but don't try to delete from the selectedItems list nor decrease selection count. The copied item is of course not in there.

        #most of the time this is ["session"]
        self.clipboard[cliplink] = tmpclipboard
        self.linkboard[cliplink] = tmplinkboard


        session.workspace.movement.head() #return to the beginning, resets all cursors.
        session.workspace.movement.activeCursor = start[0]
        session.workspace.movement.toIndex(start[1])

    def generatorClipboard(self, cliplink):
        """Return a generator clipboard. Ignores all track breaks"""
        return (obj
                for track  in self.clipboard[cliplink]
                for obj in track)

    def generatorLinkboard(self, cliplink):
        """Return a generator clipboard. Ignores all track breaks"""
        return (obj
                for track  in self.linkboard[cliplink]
                for obj in track)


    def filterClipboard(self, cliplink, test):
        """Filter a clip+link-board  so only those items remain which
        succeed in the test.
        Test is a function that must accept one parameter, the item"""
        self.clipboard[cliplink] = [[ obj
                                    for track in self.clipboard[cliplink]
                                    for obj in track if test(obj)]]

        self.linkboard[cliplink] = [[ obj
                                    for track in self.linkboard[cliplink]
                                    for obj in track if test(obj)]]

class Session(object):
    """A container to store and manage all open workspaces.
    This level of organisation is not saved or stored anywhere.
    It just exists as long the same program session is open.
    """
    def __init__(self):
        self.workspace = None  #the current file.
        self.workspaces = []
        self.substitutions = None  #hardcoded. Filled during Startup Has Container() in it. The developers add them directly to this dict. They are not saved and loaded, they just are. Dependent of the program version.
        self.configdir = os.path.expanduser("~") + "/.laborejo/" #the homedir on windows and linux for custom lilypond templates. This is the default but it can be changed after program start since all functions like lilypond export re-ask for this value every time.

    def new(self):
        """new workspace/file. Makes the new one active."""
        new = ExportContainer() #also creates the first movement with the first track with its appending position items.
        self.workspaces.append(new)
        self.workspace = new
        self.workspace.parentSession = self #an ExportContainer has the session as temporary parent. Currently only used to access the current homedir.
        return new

    def load(self, loadedWorkspace):
        """create a new workspace, but load existing data from another
        workspace to it. It does not matter how the workspace came
        to existence. Most likely through the api and saveload module.
        This means load does not close the current active workspace.
        The old one stays open, but not active.
        Achtung. It is technically possible to open from the same
        filepath twice. Since we operate completely in RAM these two
        instances are not connected except on save and load through
        a filename.
        Lets keep it in and see where this leads to in the real world.
        The problems that arise from duplicate open files are GUI ones.
        A script might find a nice way to handle the same file twice"""
        self.workspaces.append(loadedWorkspace)
        self.workspace = loadedWorkspace
        self.workspace.parentSession = self #an ExportContainer has the session as temporary parent. Currently only used to access the current homedir.
        return loadedWorkspace

    def close(self):
        """Delete the current active workspace.
        Make the last added workspace active.
        If there is none do nothing."""
        if self.workspace:
            self.workspaces.remove(self.workspace)
            if self.workspaces:
                self.workspace = self.workspaces[-1]
            else:
                self.workspace = None
            return self.workspace
        else:
            return False

